from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.cogshop1.data_classes import (
    ProcessingLog,
    ProcessingLogApply,
    ProcessingLogApplyList,
    ProcessingLogFields,
    ProcessingLogList,
    ProcessingLogTextFields,
)
from cognite.powerops.client._generated.cogshop1.data_classes._processing_log import _PROCESSINGLOG_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, Aggregations, TypeAPI


class ProcessingLogAPI(TypeAPI[ProcessingLog, ProcessingLogApply, ProcessingLogList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ProcessingLog,
            class_apply_type=ProcessingLogApply,
            class_list=ProcessingLogList,
        )
        self._view_id = view_id

    def apply(
        self, processing_log: ProcessingLogApply | Sequence[ProcessingLogApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(processing_log, ProcessingLogApply):
            instances = processing_log.to_instances_apply()
        else:
            instances = ProcessingLogApplyList(processing_log).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="cogShop") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ProcessingLog: ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ProcessingLogList: ...

    def retrieve(self, external_id: str | Sequence[str]) -> ProcessingLog | ProcessingLogList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: ProcessingLogTextFields | Sequence[ProcessingLogTextFields] | None = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        timestamp: str | list[str] | None = None,
        timestamp_prefix: str | None = None,
        error_message: str | list[str] | None = None,
        error_message_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ProcessingLogList:
        filter_ = _create_filter(
            self._view_id,
            state,
            state_prefix,
            timestamp,
            timestamp_prefix,
            error_message,
            error_message_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _PROCESSINGLOG_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ProcessingLogFields | Sequence[ProcessingLogFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ProcessingLogTextFields | Sequence[ProcessingLogTextFields] | None = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        timestamp: str | list[str] | None = None,
        timestamp_prefix: str | None = None,
        error_message: str | list[str] | None = None,
        error_message_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ProcessingLogFields | Sequence[ProcessingLogFields] | None = None,
        group_by: ProcessingLogFields | Sequence[ProcessingLogFields] = None,
        query: str | None = None,
        search_properties: ProcessingLogTextFields | Sequence[ProcessingLogTextFields] | None = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        timestamp: str | list[str] | None = None,
        timestamp_prefix: str | None = None,
        error_message: str | list[str] | None = None,
        error_message_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ProcessingLogFields | Sequence[ProcessingLogFields] | None = None,
        group_by: ProcessingLogFields | Sequence[ProcessingLogFields] | None = None,
        query: str | None = None,
        search_property: ProcessingLogTextFields | Sequence[ProcessingLogTextFields] | None = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        timestamp: str | list[str] | None = None,
        timestamp_prefix: str | None = None,
        error_message: str | list[str] | None = None,
        error_message_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            state,
            state_prefix,
            timestamp,
            timestamp_prefix,
            error_message,
            error_message_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _PROCESSINGLOG_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ProcessingLogFields,
        interval: float,
        query: str | None = None,
        search_property: ProcessingLogTextFields | Sequence[ProcessingLogTextFields] | None = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        timestamp: str | list[str] | None = None,
        timestamp_prefix: str | None = None,
        error_message: str | list[str] | None = None,
        error_message_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            state,
            state_prefix,
            timestamp,
            timestamp_prefix,
            error_message,
            error_message_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _PROCESSINGLOG_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        timestamp: str | list[str] | None = None,
        timestamp_prefix: str | None = None,
        error_message: str | list[str] | None = None,
        error_message_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ProcessingLogList:
        filter_ = _create_filter(
            self._view_id,
            state,
            state_prefix,
            timestamp,
            timestamp_prefix,
            error_message,
            error_message_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    state: str | list[str] | None = None,
    state_prefix: str | None = None,
    timestamp: str | list[str] | None = None,
    timestamp_prefix: str | None = None,
    error_message: str | list[str] | None = None,
    error_message_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if state and isinstance(state, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("state"), value=state))
    if state and isinstance(state, list):
        filters.append(dm.filters.In(view_id.as_property_ref("state"), values=state))
    if state_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("state"), value=state_prefix))
    if timestamp and isinstance(timestamp, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timestamp"), value=timestamp))
    if timestamp and isinstance(timestamp, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timestamp"), values=timestamp))
    if timestamp_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timestamp"), value=timestamp_prefix))
    if error_message and isinstance(error_message, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("errorMessage"), value=error_message))
    if error_message and isinstance(error_message, list):
        filters.append(dm.filters.In(view_id.as_property_ref("errorMessage"), values=error_message))
    if error_message_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("errorMessage"), value=error_message_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)
from ._bid_method import BidMethod, BidMethodWrite

if TYPE_CHECKING:
    from ._shop_price_scenario import SHOPPriceScenario, SHOPPriceScenarioGraphQL, SHOPPriceScenarioWrite


__all__ = [
    "SHOPMultiScenarioMethod",
    "SHOPMultiScenarioMethodWrite",
    "SHOPMultiScenarioMethodApply",
    "SHOPMultiScenarioMethodList",
    "SHOPMultiScenarioMethodWriteList",
    "SHOPMultiScenarioMethodApplyList",
    "SHOPMultiScenarioMethodFields",
    "SHOPMultiScenarioMethodTextFields",
]


SHOPMultiScenarioMethodTextFields = Literal["name", "shop_cases"]
SHOPMultiScenarioMethodFields = Literal["name", "shop_cases"]

_SHOPMULTISCENARIOMETHOD_PROPERTIES_BY_FIELD = {
    "name": "name",
    "shop_cases": "shopCases",
}


class SHOPMultiScenarioMethodGraphQL(GraphQLCore):
    """This represents the reading version of shop multi scenario method, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop multi scenario method.
        data_record: The data record of the shop multi scenario method node.
        name: Name for the BidMethod
        shop_cases: The shop case field.
        price_scenarios: An array of scenarios for this bid method.
    """

    view_id = dm.ViewId("power-ops-day-ahead-bid", "SHOPMultiScenarioMethod", "1")
    name: Optional[str] = None
    shop_cases: Optional[list[str]] = Field(None, alias="shopCases")
    price_scenarios: Optional[list[SHOPPriceScenarioGraphQL]] = Field(default=None, repr=False, alias="priceScenarios")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("price_scenarios", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> SHOPMultiScenarioMethod:
        """Convert this GraphQL format of shop multi scenario method to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return SHOPMultiScenarioMethod(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            shop_cases=self.shop_cases,
            price_scenarios=[
                price_scenario.as_read() if isinstance(price_scenario, GraphQLCore) else price_scenario
                for price_scenario in self.price_scenarios or []
            ],
        )

    def as_write(self) -> SHOPMultiScenarioMethodWrite:
        """Convert this GraphQL format of shop multi scenario method to the writing format."""
        return SHOPMultiScenarioMethodWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            shop_cases=self.shop_cases,
            price_scenarios=[
                price_scenario.as_write() if isinstance(price_scenario, DomainModel) else price_scenario
                for price_scenario in self.price_scenarios or []
            ],
        )


class SHOPMultiScenarioMethod(BidMethod):
    """This represents the reading version of shop multi scenario method.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop multi scenario method.
        data_record: The data record of the shop multi scenario method node.
        name: Name for the BidMethod
        shop_cases: The shop case field.
        price_scenarios: An array of scenarios for this bid method.
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "DayAheadSHOPMultiScenarioMethod"
    )
    shop_cases: Optional[list[str]] = Field(None, alias="shopCases")
    price_scenarios: Union[list[SHOPPriceScenario], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="priceScenarios"
    )

    def as_write(self) -> SHOPMultiScenarioMethodWrite:
        """Convert this read version of shop multi scenario method to the writing version."""
        return SHOPMultiScenarioMethodWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            shop_cases=self.shop_cases,
            price_scenarios=[
                price_scenario.as_write() if isinstance(price_scenario, DomainModel) else price_scenario
                for price_scenario in self.price_scenarios or []
            ],
        )

    def as_apply(self) -> SHOPMultiScenarioMethodWrite:
        """Convert this read version of shop multi scenario method to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPMultiScenarioMethodWrite(BidMethodWrite):
    """This represents the writing version of shop multi scenario method.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop multi scenario method.
        data_record: The data record of the shop multi scenario method node.
        name: Name for the BidMethod
        shop_cases: The shop case field.
        price_scenarios: An array of scenarios for this bid method.
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "DayAheadSHOPMultiScenarioMethod"
    )
    shop_cases: Optional[list[str]] = Field(None, alias="shopCases")
    price_scenarios: Union[list[SHOPPriceScenarioWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="priceScenarios"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            SHOPMultiScenarioMethod, dm.ViewId("power-ops-day-ahead-bid", "SHOPMultiScenarioMethod", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.shop_cases is not None or write_none:
            properties["shopCases"] = self.shop_cases

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power-ops-types", "PriceScenario")
        for price_scenario in self.price_scenarios or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=price_scenario,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        return resources


class SHOPMultiScenarioMethodApply(SHOPMultiScenarioMethodWrite):
    def __new__(cls, *args, **kwargs) -> SHOPMultiScenarioMethodApply:
        warnings.warn(
            "SHOPMultiScenarioMethodApply is deprecated and will be removed in v1.0. Use SHOPMultiScenarioMethodWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "SHOPMultiScenarioMethod.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class SHOPMultiScenarioMethodList(DomainModelList[SHOPMultiScenarioMethod]):
    """List of shop multi scenario methods in the read version."""

    _INSTANCE = SHOPMultiScenarioMethod

    def as_write(self) -> SHOPMultiScenarioMethodWriteList:
        """Convert these read versions of shop multi scenario method to the writing versions."""
        return SHOPMultiScenarioMethodWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> SHOPMultiScenarioMethodWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPMultiScenarioMethodWriteList(DomainModelWriteList[SHOPMultiScenarioMethodWrite]):
    """List of shop multi scenario methods in the writing version."""

    _INSTANCE = SHOPMultiScenarioMethodWrite


class SHOPMultiScenarioMethodApplyList(SHOPMultiScenarioMethodWriteList): ...


def _create_shop_multi_scenario_method_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, ClassVar, Literal,  no_type_check, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._alert import Alert, AlertGraphQL, AlertWrite
    from ._shop_case import ShopCase, ShopCaseGraphQL, ShopCaseWrite
    from ._shop_time_series import ShopTimeSeries, ShopTimeSeriesGraphQL, ShopTimeSeriesWrite


__all__ = [
    "ShopResult",
    "ShopResultWrite",
    "ShopResultApply",
    "ShopResultList",
    "ShopResultWriteList",
    "ShopResultApplyList",
    "ShopResultFields",
    "ShopResultTextFields",
    "ShopResultGraphQL",
]


ShopResultTextFields = Literal["pre_run", "post_run", "messages", "cplex_logs"]
ShopResultFields = Literal["objective_value", "pre_run", "post_run", "messages", "cplex_logs"]

_SHOPRESULT_PROPERTIES_BY_FIELD = {
    "objective_value": "objectiveValue",
    "pre_run": "preRun",
    "post_run": "postRun",
    "messages": "messages",
    "cplex_logs": "cplexLogs",
}

class ShopResultGraphQL(GraphQLCore):
    """This represents the reading version of shop result, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop result.
        data_record: The data record of the shop result node.
        case: The case that was used to produce this result
        objective_value: The sequence of the objective function
        pre_run: The pre-run data for the SHOP run
        post_run: The post-run data for the SHOP run
        messages: The messages from the SHOP run
        cplex_logs: The logs from CPLEX
        alerts: An array of calculation level Alerts.
        output_time_series: TODO
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopResult", "1")
    case: Optional[ShopCaseGraphQL] = Field(default=None, repr=False)
    objective_value: Optional[dict] = Field(None, alias="objectiveValue")
    pre_run: Union[dict, None] = Field(None, alias="preRun")
    post_run: Union[dict, None] = Field(None, alias="postRun")
    messages: Union[dict, None] = None
    cplex_logs: Union[dict, None] = Field(None, alias="cplexLogs")
    alerts: Optional[list[AlertGraphQL]] = Field(default=None, repr=False)
    output_time_series: Optional[list[ShopTimeSeriesGraphQL]] = Field(default=None, repr=False, alias="outputTimeSeries")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values
    @field_validator("case", "alerts", "output_time_series", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> ShopResult:
        """Convert this GraphQL format of shop result to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return ShopResult(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            case=self.case.as_read() if isinstance(self.case, GraphQLCore) else self.case,
            objective_value=self.objective_value,
            pre_run=self.pre_run["externalId"] if self.pre_run and "externalId" in self.pre_run else None,
            post_run=self.post_run["externalId"] if self.post_run and "externalId" in self.post_run else None,
            messages=self.messages["externalId"] if self.messages and "externalId" in self.messages else None,
            cplex_logs=self.cplex_logs["externalId"] if self.cplex_logs and "externalId" in self.cplex_logs else None,
            alerts=[alert.as_read() for alert in self.alerts or []],
            output_time_series=[output_time_series.as_read() for output_time_series in self.output_time_series or []],
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> ShopResultWrite:
        """Convert this GraphQL format of shop result to the writing format."""
        return ShopResultWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            case=self.case.as_write() if isinstance(self.case, GraphQLCore) else self.case,
            objective_value=self.objective_value,
            pre_run=self.pre_run["externalId"] if self.pre_run and "externalId" in self.pre_run else None,
            post_run=self.post_run["externalId"] if self.post_run and "externalId" in self.post_run else None,
            messages=self.messages["externalId"] if self.messages and "externalId" in self.messages else None,
            cplex_logs=self.cplex_logs["externalId"] if self.cplex_logs and "externalId" in self.cplex_logs else None,
            alerts=[alert.as_write() for alert in self.alerts or []],
            output_time_series=[output_time_series.as_write() for output_time_series in self.output_time_series or []],
        )


class ShopResult(DomainModel):
    """This represents the reading version of shop result.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop result.
        data_record: The data record of the shop result node.
        case: The case that was used to produce this result
        objective_value: The sequence of the objective function
        pre_run: The pre-run data for the SHOP run
        post_run: The post-run data for the SHOP run
        messages: The messages from the SHOP run
        cplex_logs: The logs from CPLEX
        alerts: An array of calculation level Alerts.
        output_time_series: TODO
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopResult", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    case: Union[ShopCase, str, dm.NodeId, None] = Field(default=None, repr=False)
    objective_value: Optional[dict] = Field(None, alias="objectiveValue")
    pre_run: Union[str, None] = Field(None, alias="preRun")
    post_run: Union[str, None] = Field(None, alias="postRun")
    messages: Union[str, None] = None
    cplex_logs: Union[str, None] = Field(None, alias="cplexLogs")
    alerts: Optional[list[Union[Alert, str, dm.NodeId]]] = Field(default=None, repr=False)
    output_time_series: Optional[list[Union[ShopTimeSeries, str, dm.NodeId]]] = Field(default=None, repr=False, alias="outputTimeSeries")

    def as_write(self) -> ShopResultWrite:
        """Convert this read version of shop result to the writing version."""
        return ShopResultWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            case=self.case.as_write() if isinstance(self.case, DomainModel) else self.case,
            objective_value=self.objective_value,
            pre_run=self.pre_run,
            post_run=self.post_run,
            messages=self.messages,
            cplex_logs=self.cplex_logs,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            output_time_series=[output_time_series.as_write() if isinstance(output_time_series, DomainModel) else output_time_series for output_time_series in self.output_time_series or []],
        )

    def as_apply(self) -> ShopResultWrite:
        """Convert this read version of shop result to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopResultWrite(DomainModelWrite):
    """This represents the writing version of shop result.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop result.
        data_record: The data record of the shop result node.
        case: The case that was used to produce this result
        objective_value: The sequence of the objective function
        pre_run: The pre-run data for the SHOP run
        post_run: The post-run data for the SHOP run
        messages: The messages from the SHOP run
        cplex_logs: The logs from CPLEX
        alerts: An array of calculation level Alerts.
        output_time_series: TODO
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopResult", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    case: Union[ShopCaseWrite, str, dm.NodeId, None] = Field(default=None, repr=False)
    objective_value: Optional[dict] = Field(None, alias="objectiveValue")
    pre_run: Union[str, None] = Field(None, alias="preRun")
    post_run: Union[str, None] = Field(None, alias="postRun")
    messages: Union[str, None] = None
    cplex_logs: Union[str, None] = Field(None, alias="cplexLogs")
    alerts: Optional[list[Union[AlertWrite, str, dm.NodeId]]] = Field(default=None, repr=False)
    output_time_series: Optional[list[Union[ShopTimeSeriesWrite, str, dm.NodeId]]] = Field(default=None, repr=False, alias="outputTimeSeries")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.case is not None:
            properties["case"] = {
                "space":  self.space if isinstance(self.case, str) else self.case.space,
                "externalId": self.case if isinstance(self.case, str) else self.case.external_id,
            }

        if self.objective_value is not None or write_none:
            properties["objectiveValue"] = self.objective_value

        if self.pre_run is not None or write_none:
            properties["preRun"] = self.pre_run

        if self.post_run is not None or write_none:
            properties["postRun"] = self.post_run

        if self.messages is not None or write_none:
            properties["messages"] = self.messages

        if self.cplex_logs is not None or write_none:
            properties["cplexLogs"] = self.cplex_logs


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        edge_type = dm.DirectRelationReference("power_ops_types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=alert,
                edge_type=edge_type,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power_ops_types", "ShopResult.outputTimeSeries")
        for output_time_series in self.output_time_series or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=output_time_series,
                edge_type=edge_type,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.case, DomainModelWrite):
            other_resources = self.case._to_instances_write(cache)
            resources.extend(other_resources)

        return resources


class ShopResultApply(ShopResultWrite):
    def __new__(cls, *args, **kwargs) -> ShopResultApply:
        warnings.warn(
            "ShopResultApply is deprecated and will be removed in v1.0. Use ShopResultWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ShopResult.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ShopResultList(DomainModelList[ShopResult]):
    """List of shop results in the read version."""

    _INSTANCE = ShopResult

    def as_write(self) -> ShopResultWriteList:
        """Convert these read versions of shop result to the writing versions."""
        return ShopResultWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ShopResultWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopResultWriteList(DomainModelWriteList[ShopResultWrite]):
    """List of shop results in the writing version."""

    _INSTANCE = ShopResultWrite

class ShopResultApplyList(ShopResultWriteList): ...



def _create_shop_result_filter(
    view_id: dm.ViewId,
    case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if case and isinstance(case, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("case"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": case}))
    if case and isinstance(case, tuple):
        filters.append(dm.filters.Equals(view_id.as_property_ref("case"), value={"space": case[0], "externalId": case[1]}))
    if case and isinstance(case, list) and isinstance(case[0], str):
        filters.append(dm.filters.In(view_id.as_property_ref("case"), values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in case]))
    if case and isinstance(case, list) and isinstance(case[0], tuple):
        filters.append(dm.filters.In(view_id.as_property_ref("case"), values=[{"space": item[0], "externalId": item[1]} for item in case]))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, ClassVar, Literal,  no_type_check, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._date_specification import DateSpecification, DateSpecificationGraphQL, DateSpecificationWrite
    from ._market_configuration import MarketConfiguration, MarketConfigurationGraphQL, MarketConfigurationWrite
    from ._partial_bid_configuration import PartialBidConfiguration, PartialBidConfigurationGraphQL, PartialBidConfigurationWrite
    from ._price_area_day_ahead import PriceAreaDayAhead, PriceAreaDayAheadGraphQL, PriceAreaDayAheadWrite


__all__ = [
    "BidConfigurationDayAhead",
    "BidConfigurationDayAheadWrite",
    "BidConfigurationDayAheadApply",
    "BidConfigurationDayAheadList",
    "BidConfigurationDayAheadWriteList",
    "BidConfigurationDayAheadApplyList",
    "BidConfigurationDayAheadFields",
    "BidConfigurationDayAheadTextFields",
    "BidConfigurationDayAheadGraphQL",
]


BidConfigurationDayAheadTextFields = Literal["name"]
BidConfigurationDayAheadFields = Literal["name"]

_BIDCONFIGURATIONDAYAHEAD_PROPERTIES_BY_FIELD = {
    "name": "name",
}

class BidConfigurationDayAheadGraphQL(GraphQLCore):
    """This represents the reading version of bid configuration day ahead, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid configuration day ahead.
        data_record: The data record of the bid configuration day ahead node.
        name: The name of the bid configuration
        market_configuration: The market configuration related to the bid configuration
        price_area: The price area related to the bid calculation task
        bid_date_specification: TODO description
        partials: Configuration of the partial bids that make up the total bid
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1")
    name: Optional[str] = None
    market_configuration: Optional[MarketConfigurationGraphQL] = Field(default=None, repr=False, alias="marketConfiguration")
    price_area: Optional[PriceAreaDayAheadGraphQL] = Field(default=None, repr=False, alias="priceArea")
    bid_date_specification: Optional[DateSpecificationGraphQL] = Field(default=None, repr=False, alias="bidDateSpecification")
    partials: Optional[list[PartialBidConfigurationGraphQL]] = Field(default=None, repr=False)

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values
    @field_validator("market_configuration", "price_area", "bid_date_specification", "partials", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> BidConfigurationDayAhead:
        """Convert this GraphQL format of bid configuration day ahead to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return BidConfigurationDayAhead(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            market_configuration=self.market_configuration.as_read() if isinstance(self.market_configuration, GraphQLCore) else self.market_configuration,
            price_area=self.price_area.as_read() if isinstance(self.price_area, GraphQLCore) else self.price_area,
            bid_date_specification=self.bid_date_specification.as_read() if isinstance(self.bid_date_specification, GraphQLCore) else self.bid_date_specification,
            partials=[partial.as_read() for partial in self.partials or []],
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> BidConfigurationDayAheadWrite:
        """Convert this GraphQL format of bid configuration day ahead to the writing format."""
        return BidConfigurationDayAheadWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            market_configuration=self.market_configuration.as_write() if isinstance(self.market_configuration, GraphQLCore) else self.market_configuration,
            price_area=self.price_area.as_write() if isinstance(self.price_area, GraphQLCore) else self.price_area,
            bid_date_specification=self.bid_date_specification.as_write() if isinstance(self.bid_date_specification, GraphQLCore) else self.bid_date_specification,
            partials=[partial.as_write() for partial in self.partials or []],
        )


class BidConfigurationDayAhead(DomainModel):
    """This represents the reading version of bid configuration day ahead.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid configuration day ahead.
        data_record: The data record of the bid configuration day ahead node.
        name: The name of the bid configuration
        market_configuration: The market configuration related to the bid configuration
        price_area: The price area related to the bid calculation task
        bid_date_specification: TODO description
        partials: Configuration of the partial bids that make up the total bid
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "BidConfigurationDayAhead")
    name: str
    market_configuration: Union[MarketConfiguration, str, dm.NodeId, None] = Field(default=None, repr=False, alias="marketConfiguration")
    price_area: Union[PriceAreaDayAhead, str, dm.NodeId, None] = Field(default=None, repr=False, alias="priceArea")
    bid_date_specification: Union[DateSpecification, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidDateSpecification")
    partials: Optional[list[Union[PartialBidConfiguration, str, dm.NodeId]]] = Field(default=None, repr=False)

    def as_write(self) -> BidConfigurationDayAheadWrite:
        """Convert this read version of bid configuration day ahead to the writing version."""
        return BidConfigurationDayAheadWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            market_configuration=self.market_configuration.as_write() if isinstance(self.market_configuration, DomainModel) else self.market_configuration,
            price_area=self.price_area.as_write() if isinstance(self.price_area, DomainModel) else self.price_area,
            bid_date_specification=self.bid_date_specification.as_write() if isinstance(self.bid_date_specification, DomainModel) else self.bid_date_specification,
            partials=[partial.as_write() if isinstance(partial, DomainModel) else partial for partial in self.partials or []],
        )

    def as_apply(self) -> BidConfigurationDayAheadWrite:
        """Convert this read version of bid configuration day ahead to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidConfigurationDayAheadWrite(DomainModelWrite):
    """This represents the writing version of bid configuration day ahead.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid configuration day ahead.
        data_record: The data record of the bid configuration day ahead node.
        name: The name of the bid configuration
        market_configuration: The market configuration related to the bid configuration
        price_area: The price area related to the bid calculation task
        bid_date_specification: TODO description
        partials: Configuration of the partial bids that make up the total bid
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "BidConfigurationDayAhead")
    name: str
    market_configuration: Union[MarketConfigurationWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="marketConfiguration")
    price_area: Union[PriceAreaDayAheadWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="priceArea")
    bid_date_specification: Union[DateSpecificationWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidDateSpecification")
    partials: Optional[list[Union[PartialBidConfigurationWrite, str, dm.NodeId]]] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.market_configuration is not None:
            properties["marketConfiguration"] = {
                "space":  self.space if isinstance(self.market_configuration, str) else self.market_configuration.space,
                "externalId": self.market_configuration if isinstance(self.market_configuration, str) else self.market_configuration.external_id,
            }

        if self.price_area is not None:
            properties["priceArea"] = {
                "space":  self.space if isinstance(self.price_area, str) else self.price_area.space,
                "externalId": self.price_area if isinstance(self.price_area, str) else self.price_area.external_id,
            }

        if self.bid_date_specification is not None:
            properties["bidDateSpecification"] = {
                "space":  self.space if isinstance(self.bid_date_specification, str) else self.bid_date_specification.space,
                "externalId": self.bid_date_specification if isinstance(self.bid_date_specification, str) else self.bid_date_specification.external_id,
            }


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        edge_type = dm.DirectRelationReference("power_ops_types", "BidConfiguration.partials")
        for partial in self.partials or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=partial,
                edge_type=edge_type,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.market_configuration, DomainModelWrite):
            other_resources = self.market_configuration._to_instances_write(cache)
            resources.extend(other_resources)

        if isinstance(self.price_area, DomainModelWrite):
            other_resources = self.price_area._to_instances_write(cache)
            resources.extend(other_resources)

        if isinstance(self.bid_date_specification, DomainModelWrite):
            other_resources = self.bid_date_specification._to_instances_write(cache)
            resources.extend(other_resources)

        return resources


class BidConfigurationDayAheadApply(BidConfigurationDayAheadWrite):
    def __new__(cls, *args, **kwargs) -> BidConfigurationDayAheadApply:
        warnings.warn(
            "BidConfigurationDayAheadApply is deprecated and will be removed in v1.0. Use BidConfigurationDayAheadWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BidConfigurationDayAhead.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BidConfigurationDayAheadList(DomainModelList[BidConfigurationDayAhead]):
    """List of bid configuration day aheads in the read version."""

    _INSTANCE = BidConfigurationDayAhead

    def as_write(self) -> BidConfigurationDayAheadWriteList:
        """Convert these read versions of bid configuration day ahead to the writing versions."""
        return BidConfigurationDayAheadWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BidConfigurationDayAheadWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BidConfigurationDayAheadWriteList(DomainModelWriteList[BidConfigurationDayAheadWrite]):
    """List of bid configuration day aheads in the writing version."""

    _INSTANCE = BidConfigurationDayAheadWrite

class BidConfigurationDayAheadApplyList(BidConfigurationDayAheadWriteList): ...



def _create_bid_configuration_day_ahead_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    market_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    bid_date_specification: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if market_configuration and isinstance(market_configuration, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("marketConfiguration"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": market_configuration}))
    if market_configuration and isinstance(market_configuration, tuple):
        filters.append(dm.filters.Equals(view_id.as_property_ref("marketConfiguration"), value={"space": market_configuration[0], "externalId": market_configuration[1]}))
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], str):
        filters.append(dm.filters.In(view_id.as_property_ref("marketConfiguration"), values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in market_configuration]))
    if market_configuration and isinstance(market_configuration, list) and isinstance(market_configuration[0], tuple):
        filters.append(dm.filters.In(view_id.as_property_ref("marketConfiguration"), values=[{"space": item[0], "externalId": item[1]} for item in market_configuration]))
    if price_area and isinstance(price_area, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("priceArea"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": price_area}))
    if price_area and isinstance(price_area, tuple):
        filters.append(dm.filters.Equals(view_id.as_property_ref("priceArea"), value={"space": price_area[0], "externalId": price_area[1]}))
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], str):
        filters.append(dm.filters.In(view_id.as_property_ref("priceArea"), values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in price_area]))
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], tuple):
        filters.append(dm.filters.In(view_id.as_property_ref("priceArea"), values=[{"space": item[0], "externalId": item[1]} for item in price_area]))
    if bid_date_specification and isinstance(bid_date_specification, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("bidDateSpecification"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": bid_date_specification}))
    if bid_date_specification and isinstance(bid_date_specification, tuple):
        filters.append(dm.filters.Equals(view_id.as_property_ref("bidDateSpecification"), value={"space": bid_date_specification[0], "externalId": bid_date_specification[1]}))
    if bid_date_specification and isinstance(bid_date_specification, list) and isinstance(bid_date_specification[0], str):
        filters.append(dm.filters.In(view_id.as_property_ref("bidDateSpecification"), values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in bid_date_specification]))
    if bid_date_specification and isinstance(bid_date_specification, list) and isinstance(bid_date_specification[0], tuple):
        filters.append(dm.filters.In(view_id.as_property_ref("bidDateSpecification"), values=[{"space": item[0], "externalId": item[1]} for item in bid_date_specification]))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import Any, ClassVar, Literal, no_type_check, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)


__all__ = [
    "DateSpecification",
    "DateSpecificationWrite",
    "DateSpecificationApply",
    "DateSpecificationList",
    "DateSpecificationWriteList",
    "DateSpecificationApplyList",
    "DateSpecificationFields",
    "DateSpecificationTextFields",
    "DateSpecificationGraphQL",
]


DateSpecificationTextFields = Literal["name", "processing_timezone", "resulting_timezone", "floor_frame"]
DateSpecificationFields = Literal["name", "processing_timezone", "resulting_timezone", "floor_frame", "shift_definition"]

_DATESPECIFICATION_PROPERTIES_BY_FIELD = {
    "name": "name",
    "processing_timezone": "processingTimezone",
    "resulting_timezone": "resultingTimezone",
    "floor_frame": "floorFrame",
    "shift_definition": "shiftDefinition",
}

class DateSpecificationGraphQL(GraphQLCore):
    """This represents the reading version of date specification, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the date specification.
        data_record: The data record of the date specification node.
        name: TODO description
        processing_timezone: TODO description
        resulting_timezone: TODO description
        floor_frame: TODO description
        shift_definition: TODO description
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "DateSpecification", "1")
    name: Optional[str] = None
    processing_timezone: Optional[str] = Field(None, alias="processingTimezone")
    resulting_timezone: Optional[str] = Field(None, alias="resultingTimezone")
    floor_frame: Optional[str] = Field(None, alias="floorFrame")
    shift_definition: Optional[dict] = Field(None, alias="shiftDefinition")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> DateSpecification:
        """Convert this GraphQL format of date specification to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return DateSpecification(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            processing_timezone=self.processing_timezone,
            resulting_timezone=self.resulting_timezone,
            floor_frame=self.floor_frame,
            shift_definition=self.shift_definition,
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> DateSpecificationWrite:
        """Convert this GraphQL format of date specification to the writing format."""
        return DateSpecificationWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            processing_timezone=self.processing_timezone,
            resulting_timezone=self.resulting_timezone,
            floor_frame=self.floor_frame,
            shift_definition=self.shift_definition,
        )


class DateSpecification(DomainModel):
    """This represents the reading version of date specification.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the date specification.
        data_record: The data record of the date specification node.
        name: TODO description
        processing_timezone: TODO description
        resulting_timezone: TODO description
        floor_frame: TODO description
        shift_definition: TODO description
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "DateSpecification", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "DateSpecification")
    name: str
    processing_timezone: Optional[str] = Field(None, alias="processingTimezone")
    resulting_timezone: Optional[str] = Field(None, alias="resultingTimezone")
    floor_frame: Optional[str] = Field(None, alias="floorFrame")
    shift_definition: Optional[dict] = Field(None, alias="shiftDefinition")

    def as_write(self) -> DateSpecificationWrite:
        """Convert this read version of date specification to the writing version."""
        return DateSpecificationWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            processing_timezone=self.processing_timezone,
            resulting_timezone=self.resulting_timezone,
            floor_frame=self.floor_frame,
            shift_definition=self.shift_definition,
        )

    def as_apply(self) -> DateSpecificationWrite:
        """Convert this read version of date specification to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class DateSpecificationWrite(DomainModelWrite):
    """This represents the writing version of date specification.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the date specification.
        data_record: The data record of the date specification node.
        name: TODO description
        processing_timezone: TODO description
        resulting_timezone: TODO description
        floor_frame: TODO description
        shift_definition: TODO description
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "DateSpecification", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "DateSpecification")
    name: str
    processing_timezone: Optional[str] = Field("UTC", alias="processingTimezone")
    resulting_timezone: Optional[str] = Field("UTC", alias="resultingTimezone")
    floor_frame: Optional[str] = Field("day", alias="floorFrame")
    shift_definition: Optional[dict] = Field(None, alias="shiftDefinition")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.processing_timezone is not None or write_none:
            properties["processingTimezone"] = self.processing_timezone

        if self.resulting_timezone is not None or write_none:
            properties["resultingTimezone"] = self.resulting_timezone

        if self.floor_frame is not None or write_none:
            properties["floorFrame"] = self.floor_frame

        if self.shift_definition is not None or write_none:
            properties["shiftDefinition"] = self.shift_definition


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        return resources


class DateSpecificationApply(DateSpecificationWrite):
    def __new__(cls, *args, **kwargs) -> DateSpecificationApply:
        warnings.warn(
            "DateSpecificationApply is deprecated and will be removed in v1.0. Use DateSpecificationWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "DateSpecification.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class DateSpecificationList(DomainModelList[DateSpecification]):
    """List of date specifications in the read version."""

    _INSTANCE = DateSpecification

    def as_write(self) -> DateSpecificationWriteList:
        """Convert these read versions of date specification to the writing versions."""
        return DateSpecificationWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> DateSpecificationWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class DateSpecificationWriteList(DomainModelWriteList[DateSpecificationWrite]):
    """List of date specifications in the writing version."""

    _INSTANCE = DateSpecificationWrite

class DateSpecificationApplyList(DateSpecificationWriteList): ...



def _create_date_specification_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    processing_timezone: str | list[str] | None = None,
    processing_timezone_prefix: str | None = None,
    resulting_timezone: str | list[str] | None = None,
    resulting_timezone_prefix: str | None = None,
    floor_frame: str | list[str] | None = None,
    floor_frame_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(processing_timezone, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("processingTimezone"), value=processing_timezone))
    if processing_timezone and isinstance(processing_timezone, list):
        filters.append(dm.filters.In(view_id.as_property_ref("processingTimezone"), values=processing_timezone))
    if processing_timezone_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("processingTimezone"), value=processing_timezone_prefix))
    if isinstance(resulting_timezone, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("resultingTimezone"), value=resulting_timezone))
    if resulting_timezone and isinstance(resulting_timezone, list):
        filters.append(dm.filters.In(view_id.as_property_ref("resultingTimezone"), values=resulting_timezone))
    if resulting_timezone_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("resultingTimezone"), value=resulting_timezone_prefix))
    if isinstance(floor_frame, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("floorFrame"), value=floor_frame))
    if floor_frame and isinstance(floor_frame, list):
        filters.append(dm.filters.In(view_id.as_property_ref("floorFrame"), values=floor_frame))
    if floor_frame_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("floorFrame"), value=floor_frame_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import Any, ClassVar, Literal, no_type_check, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)


__all__ = [
    "MarketConfiguration",
    "MarketConfigurationWrite",
    "MarketConfigurationApply",
    "MarketConfigurationList",
    "MarketConfigurationWriteList",
    "MarketConfigurationApplyList",
    "MarketConfigurationFields",
    "MarketConfigurationTextFields",
    "MarketConfigurationGraphQL",
]


MarketConfigurationTextFields = Literal["name", "timezone", "price_unit", "time_unit"]
MarketConfigurationFields = Literal["name", "max_price", "min_price", "timezone", "price_unit", "price_steps", "tick_size", "time_unit", "trade_lot"]

_MARKETCONFIGURATION_PROPERTIES_BY_FIELD = {
    "name": "name",
    "max_price": "maxPrice",
    "min_price": "minPrice",
    "timezone": "timezone",
    "price_unit": "priceUnit",
    "price_steps": "priceSteps",
    "tick_size": "tickSize",
    "time_unit": "timeUnit",
    "trade_lot": "tradeLot",
}

class MarketConfigurationGraphQL(GraphQLCore):
    """This represents the reading version of market configuration, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the market configuration.
        data_record: The data record of the market configuration node.
        name: The name of the market
        max_price: The highest price allowed
        min_price: The lowest price allowed
        timezone: The timezone field.
        price_unit: Unit of measurement for the price ('EUR/MWh')
        price_steps: The maximum number of price steps
        tick_size: 'Granularity' of the price; tick size = 0.1 means that prices must be 'rounded to nearest 0.1' (i. e. 66.43 is not allowed, but 66.4 is)
        time_unit: The time unit ('1h')
        trade_lot: 'Granularity' of the volumes; trade lot = 0.2 means that volumes must be 'rounded to nearest 0.2' (i. e. 66.5 is not allowed, but 66.4 is)
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "MarketConfiguration", "1")
    name: Optional[str] = None
    max_price: Optional[float] = Field(None, alias="maxPrice")
    min_price: Optional[float] = Field(None, alias="minPrice")
    timezone: Optional[str] = None
    price_unit: Optional[str] = Field(None, alias="priceUnit")
    price_steps: Optional[int] = Field(None, alias="priceSteps")
    tick_size: Optional[float] = Field(None, alias="tickSize")
    time_unit: Optional[str] = Field(None, alias="timeUnit")
    trade_lot: Optional[float] = Field(None, alias="tradeLot")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> MarketConfiguration:
        """Convert this GraphQL format of market configuration to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return MarketConfiguration(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            max_price=self.max_price,
            min_price=self.min_price,
            timezone=self.timezone,
            price_unit=self.price_unit,
            price_steps=self.price_steps,
            tick_size=self.tick_size,
            time_unit=self.time_unit,
            trade_lot=self.trade_lot,
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> MarketConfigurationWrite:
        """Convert this GraphQL format of market configuration to the writing format."""
        return MarketConfigurationWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            max_price=self.max_price,
            min_price=self.min_price,
            timezone=self.timezone,
            price_unit=self.price_unit,
            price_steps=self.price_steps,
            tick_size=self.tick_size,
            time_unit=self.time_unit,
            trade_lot=self.trade_lot,
        )


class MarketConfiguration(DomainModel):
    """This represents the reading version of market configuration.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the market configuration.
        data_record: The data record of the market configuration node.
        name: The name of the market
        max_price: The highest price allowed
        min_price: The lowest price allowed
        timezone: The timezone field.
        price_unit: Unit of measurement for the price ('EUR/MWh')
        price_steps: The maximum number of price steps
        tick_size: 'Granularity' of the price; tick size = 0.1 means that prices must be 'rounded to nearest 0.1' (i. e. 66.43 is not allowed, but 66.4 is)
        time_unit: The time unit ('1h')
        trade_lot: 'Granularity' of the volumes; trade lot = 0.2 means that volumes must be 'rounded to nearest 0.2' (i. e. 66.5 is not allowed, but 66.4 is)
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "MarketConfiguration", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "MarketConfiguration")
    name: str
    max_price: float = Field(alias="maxPrice")
    min_price: float = Field(alias="minPrice")
    timezone: str
    price_unit: str = Field(alias="priceUnit")
    price_steps: int = Field(alias="priceSteps")
    tick_size: float = Field(alias="tickSize")
    time_unit: str = Field(alias="timeUnit")
    trade_lot: float = Field(alias="tradeLot")

    def as_write(self) -> MarketConfigurationWrite:
        """Convert this read version of market configuration to the writing version."""
        return MarketConfigurationWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            max_price=self.max_price,
            min_price=self.min_price,
            timezone=self.timezone,
            price_unit=self.price_unit,
            price_steps=self.price_steps,
            tick_size=self.tick_size,
            time_unit=self.time_unit,
            trade_lot=self.trade_lot,
        )

    def as_apply(self) -> MarketConfigurationWrite:
        """Convert this read version of market configuration to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class MarketConfigurationWrite(DomainModelWrite):
    """This represents the writing version of market configuration.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the market configuration.
        data_record: The data record of the market configuration node.
        name: The name of the market
        max_price: The highest price allowed
        min_price: The lowest price allowed
        timezone: The timezone field.
        price_unit: Unit of measurement for the price ('EUR/MWh')
        price_steps: The maximum number of price steps
        tick_size: 'Granularity' of the price; tick size = 0.1 means that prices must be 'rounded to nearest 0.1' (i. e. 66.43 is not allowed, but 66.4 is)
        time_unit: The time unit ('1h')
        trade_lot: 'Granularity' of the volumes; trade lot = 0.2 means that volumes must be 'rounded to nearest 0.2' (i. e. 66.5 is not allowed, but 66.4 is)
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "MarketConfiguration", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "MarketConfiguration")
    name: str
    max_price: float = Field(alias="maxPrice")
    min_price: float = Field(alias="minPrice")
    timezone: str
    price_unit: str = Field(alias="priceUnit")
    price_steps: int = Field(alias="priceSteps")
    tick_size: float = Field(alias="tickSize")
    time_unit: str = Field(alias="timeUnit")
    trade_lot: float = Field(alias="tradeLot")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.max_price is not None:
            properties["maxPrice"] = self.max_price

        if self.min_price is not None:
            properties["minPrice"] = self.min_price

        if self.timezone is not None:
            properties["timezone"] = self.timezone

        if self.price_unit is not None:
            properties["priceUnit"] = self.price_unit

        if self.price_steps is not None:
            properties["priceSteps"] = self.price_steps

        if self.tick_size is not None:
            properties["tickSize"] = self.tick_size

        if self.time_unit is not None:
            properties["timeUnit"] = self.time_unit

        if self.trade_lot is not None:
            properties["tradeLot"] = self.trade_lot


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        return resources


class MarketConfigurationApply(MarketConfigurationWrite):
    def __new__(cls, *args, **kwargs) -> MarketConfigurationApply:
        warnings.warn(
            "MarketConfigurationApply is deprecated and will be removed in v1.0. Use MarketConfigurationWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "MarketConfiguration.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class MarketConfigurationList(DomainModelList[MarketConfiguration]):
    """List of market configurations in the read version."""

    _INSTANCE = MarketConfiguration

    def as_write(self) -> MarketConfigurationWriteList:
        """Convert these read versions of market configuration to the writing versions."""
        return MarketConfigurationWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> MarketConfigurationWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class MarketConfigurationWriteList(DomainModelWriteList[MarketConfigurationWrite]):
    """List of market configurations in the writing version."""

    _INSTANCE = MarketConfigurationWrite

class MarketConfigurationApplyList(MarketConfigurationWriteList): ...



def _create_market_configuration_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    min_max_price: float | None = None,
    max_max_price: float | None = None,
    min_min_price: float | None = None,
    max_min_price: float | None = None,
    timezone: str | list[str] | None = None,
    timezone_prefix: str | None = None,
    price_unit: str | list[str] | None = None,
    price_unit_prefix: str | None = None,
    min_price_steps: int | None = None,
    max_price_steps: int | None = None,
    min_tick_size: float | None = None,
    max_tick_size: float | None = None,
    time_unit: str | list[str] | None = None,
    time_unit_prefix: str | None = None,
    min_trade_lot: float | None = None,
    max_trade_lot: float | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if min_max_price is not None or max_max_price is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("maxPrice"), gte=min_max_price, lte=max_max_price))
    if min_min_price is not None or max_min_price is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("minPrice"), gte=min_min_price, lte=max_min_price))
    if isinstance(timezone, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timezone"), value=timezone))
    if timezone and isinstance(timezone, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timezone"), values=timezone))
    if timezone_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timezone"), value=timezone_prefix))
    if isinstance(price_unit, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("priceUnit"), value=price_unit))
    if price_unit and isinstance(price_unit, list):
        filters.append(dm.filters.In(view_id.as_property_ref("priceUnit"), values=price_unit))
    if price_unit_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("priceUnit"), value=price_unit_prefix))
    if min_price_steps is not None or max_price_steps is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("priceSteps"), gte=min_price_steps, lte=max_price_steps))
    if min_tick_size is not None or max_tick_size is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("tickSize"), gte=min_tick_size, lte=max_tick_size))
    if isinstance(time_unit, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timeUnit"), value=time_unit))
    if time_unit and isinstance(time_unit, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timeUnit"), values=time_unit))
    if time_unit_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timeUnit"), value=time_unit_prefix))
    if min_trade_lot is not None or max_trade_lot is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("tradeLot"), gte=min_trade_lot, lte=max_trade_lot))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import abc
import logging
from collections.abc import Sequence
from datetime import datetime
from pathlib import Path
from typing import Generic, TypeVar, Union

import yaml
from cognite.client.data_classes import FileMetadata

from cognite.powerops.utils.helpers import get_dict_dot_keys, is_time_series_dict

logger = logging.getLogger(__name__)


FileContentTypeT = TypeVar("FileContentTypeT", bound=Union[str, dict])


class SHOPResultFile(abc.ABC, Generic[FileContentTypeT]):
    """Base class for handling a results file from SHOP."""

    def __init__(
        self,
        content: FileContentTypeT,
        file_metadata: FileMetadata = None,
        encoding="utf-8",
    ) -> None:
        self._file_metadata = file_metadata
        self._encoding = encoding
        super().__init__()
        self.data: FileContentTypeT = content

    @property
    def encoding(self) -> str:
        """
        Encoding of the file.

        Returns:
            Encoding of the file.
        """
        return self._encoding

    @property
    def external_id(self):
        """
        External ID of the file.

        Returns:
            External ID of the file.
        """
        return self._file_metadata.external_id

    @property
    def name(self):
        """
        Name of the file.

        Returns:
            Name of the file.
        """

        return self._file_metadata.name

    def save_to_disk(self, dir_path: str = "") -> Path:
        """Save file to local filesystem."""
        file_path = (Path(dir_path) or Path.cwd()) / self._file_metadata.external_id
        Path(dir_path).mkdir(parents=True, exist_ok=True)
        with file_path.open("w", encoding=self.encoding) as out_file:
            out_file.write(self.file_content)
        return file_path

    @property
    def file_content(self) -> str:
        """Data encoded for writing to local filesystem."""
        raise NotImplementedError()


class SHOPLogFile(SHOPResultFile[str]):
    """
    Plain text result file (for SHOP messages and CPlex logs).
    """

    @property
    def file_content(self) -> str:
        """
        Data encoded for writing to local filesystem.
        """
        return self.data


class SHOPYamlFile(SHOPResultFile[dict]):
    """
    Yaml-formatted results file (for post_run.yaml file created by SHOP).
    """

    @property
    def file_content(self) -> str:
        """
        Data encoded for writing to local filesystem.
        """
        return yaml.safe_dump(self.data, sort_keys=False)

    def _retrieve_time_series_dict(self, key: str) -> dict[datetime, float]:
        # key is a dot separated string of nested keys
        try:
            data = get_dict_dot_keys(self.data, key)
            if not is_time_series_dict(data):
                raise ValueError
            return data
        except KeyError:
            logger.error(f'Key "{key}" not found in {self.name}')
        except ValueError:
            logger.error(f'Data at with key "{key}" cannot be plotted as a time series')
        return {}

    def _prepare_plot_time_series(self, keys: Union[str, Sequence[str]]) -> dict:
        if isinstance(keys, str):
            keys = [keys]
        return {key: self._retrieve_time_series_dict(key) for key in keys}

    def _case_insensitive_filter_out(self, str_list: list[str], to_match: Union[str, int]) -> bool:
        """Some keys are parsed as numbers by the yaml parser"""
        return all(str(to_match).lower() != str_in_list.lower() for str_in_list in str_list)

    def find_time_series(
        self,
        matches_object_types: Sequence[str] | str = "",
        matches_object_names: Sequence[str] | str = "",
        matches_attribute_names: Sequence[str] | str = "",
    ) -> list[str]:
        """
        Find time series in the results file.

        Args:
            matches_object_types: Object types to match.
            matches_object_names: Object names to match.
            matches_attribute_names: Attribute names to match.

        Returns:
            List of dot separated strings of nested keys to the time series data.
        """
        if matches_object_types and isinstance(matches_object_types, str):
            matches_object_types = [matches_object_types]

        if matches_object_names and isinstance(matches_object_names, str):
            matches_object_names = [matches_object_names]

        if matches_attribute_names and isinstance(matches_attribute_names, str):
            matches_attribute_names = [matches_attribute_names]

        keys = []
        model = self.data["model"]
        for key1 in model:
            if matches_object_types and self._case_insensitive_filter_out(matches_object_types, key1):
                continue
            object_type: dict[str, dict] = model[key1]
            for key2, object_name in object_type.items():
                if matches_object_names and self._case_insensitive_filter_out(matches_object_names, key2):
                    continue
                for key3 in object_name:
                    if matches_attribute_names and self._case_insensitive_filter_out(matches_attribute_names, key3):
                        continue
                    attribute = object_name[key3]
                    if isinstance(attribute, dict) and all(isinstance(x, datetime) for x in attribute.keys()):
                        keys.append(f"model.{key1}.{key2}.{key3}")
        return keys



import datetime
from urllib.parse import urlparse

import requests
from cognite.client import CogniteClient

from cognite.powerops.client._generated.v1._api_client import PowerOpsModelsV1Client
from cognite.powerops.client._generated.v1.data_classes import (
    ShopCase,
    ShopCaseWrite,
    ShopFileWrite,
    ShopModelWrite,
    ShopResult,
    ShopResultList,
    ShopScenarioWrite,
)
from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE


class CogShopAPI:
    def __init__(self, cdf: CogniteClient, po: PowerOpsModelsV1Client):
        self._cdf = cdf
        self._po = po

    def _shop_url_cshaas(self) -> str:
        project = self._cdf.config.project

        cluster = urlparse(self._cdf.config.base_url).netloc.split(".", 1)[0]

        environment = ".staging" if project == "power-ops-staging" else ""

        return f"https://power-ops-api{environment}.{cluster}.cognite.ai/{project}/run-shop-as-service"

    def trigger_shop_case(self, shop_case_external_id: str):
        def auth(r: requests.PreparedRequest) -> requests.PreparedRequest:
            auth_header_name, auth_header_value = self._cdf._config.credentials.authorization_header()
            r.headers[auth_header_name] = auth_header_value
            return r

        shop_url = self._shop_url_cshaas()
        shop_body = {
            "mode": "fdm",
            "runs": [{"case_external_id": shop_case_external_id}],
        }

        response = requests.post(
            url=shop_url,
            json=shop_body,
            auth=auth,
        )
        response.raise_for_status()

    def _validate_shop_scenario_reference(
        self, shop_scenario_reference: str | ShopScenarioWrite
    ) -> str | ShopScenarioWrite:
        """Checks if the provided shop scenario is valid.
        Either it can be a Write object which is usable when uploaded or an external ID of an already existing scenario.
        """
        if isinstance(shop_scenario_reference, ShopScenarioWrite):
            # allow the write object to be used directly
            return shop_scenario_reference

        elif isinstance(shop_scenario_reference, str):
            # if the scenario reference is a string, it is assumed to be an external id
            if scenario := self._po.shop_based_day_ahead_bid_process.shop_scenario.retrieve(
                external_id=shop_scenario_reference
            ):
                return scenario.external_id

        raise ValueError(f"Invalid shop scenario reference: {shop_scenario_reference}.")

    def prepare_shop_case_with_existing_scenario(
        self,
        shop_file_list: list[tuple[str, str, bool, str]],
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        shop_scenario_reference: str | ShopScenarioWrite,
        case_external_id: str | None = None,
    ) -> ShopCaseWrite:
        """
        Prepare a SHOP case that can be written to CDF, specifying just a scenario we wish to use.

        case_external_id must be unique if provided. If it is not provided, it will be auto-generated.


        Args:
          shop_file_list: List of (file_reference, file_name, is_ascii, label) tuples,
                          in the order the files should be loaded.
             `file_reference`: external id of file in CDF.
             `file_name`: Name of the file.
             `is_ascii`: Whether the file is in ASCII format.
             `label`: Label to be added to the file, use "" if no labels
          start_time: Start of time range SHOP is optimized over. Required.
          end_time: End of time range SHOP is optimized overs. Required.
          shop_scenario_reference: An external id of an existing scenario or a `ShopScenarioWrite` instance. Required.
          case_external_id: External ID of the ShopCase instance that will be created. Optional, must be unique.

        Returns:
            ShopCaseWrite: A SHOP case that can be written to CDF
        """

        shop_files_write = [
            ShopFileWrite(
                name=file_name,
                fileReference=file_reference,
                isAscii=is_ascii,
                label=label,
                order=i + 1,  # Order is 1-indexed
            )
            for i, (file_reference, file_name, is_ascii, label) in enumerate(shop_file_list)
        ]

        case_write = ShopCaseWrite(
            start_time=start_time,
            end_time=end_time,
            scenario=self._validate_shop_scenario_reference(shop_scenario_reference),
            shop_files=shop_files_write,
            **({"external_id": case_external_id} if case_external_id else {}),
        )
        return case_write

    def prepare_shop_case(
        self,
        shop_file_list: list[tuple[str, str, bool, str]],
        shop_version: str,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        model_name: str,
        scenario_name: str,
        model_external_id: str | None = None,
        scenario_external_id: str | None = None,
        case_external_id: str | None = None,
    ) -> ShopCaseWrite:
        """
        NB! This is not the recommended way to create a SHOP case.

        Prepare a SHOP case that can be written to cdf.
        External ids must be unique. If they are not provided, they will be generated.

        In this case, `ShopScenario` as and its `ShopModel` are mostly superfluous.
        However, they are still added as nearly empty objects in order to set the SHOP version.

        Args:
          shop_file_list: List of (file_reference, file_name, is_ascii, label) tuples,
                          in the order the files should be loaded.
            `file_reference`: external id of file in CDF.
            `file_name`: Name of the file.
            `is_ascii`: Whether the file is in ASCII format.
            `label`: Label to be added to the file, use "" if no labels

          start_time: Start of time range SHOP is optimized over
          end_time: End of time range SHOP is optimized over
          shop_version: Version of SHOP to use (e.g. '16.0.2'or '15.7.0.0'). Required.

          scenario_name: Name of the scenario. Required, does not have to be unique
          model_name: Name of the model. Required, does not have to be unique
          scenario_external_id: External ID of the ShopScenario instance that will be created. Optional, must be unique
          model_external_id: External ID of the ShopModel instance that will be created. Optional, must be unique
          case_external_id: External ID of the ShopCase instance that will be created. Optional, must be unique

        Returns:
            ShopCaseWrite: A SHOP case that can be written to CDF
        """
        model_write = ShopModelWrite(
            name=model_name,
            shop_version=shop_version,
            **({"external_id": model_external_id} if model_external_id else {}),
        )

        scenario_write = ShopScenarioWrite(
            name=scenario_name,
            model=model_write,
            **({"external_id": scenario_external_id} if scenario_external_id else {}),
        )

        return self.prepare_shop_case_with_existing_scenario(
            shop_file_list=shop_file_list,
            start_time=start_time,
            end_time=end_time,
            shop_scenario_reference=scenario_write,
            case_external_id=case_external_id,
        )

    def retrieve_shop_case(self, case_external_id: str) -> ShopCase:
        """Retrieve a shop case from CDF"""
        return self._po.shop_based_day_ahead_bid_process.shop_case.retrieve(external_id=case_external_id)

    def list_shop_results_for_case(self, case_external_id: str, limit: int = 3) -> ShopResultList:
        """
        View the result of a SHOP case.
        Args:
            case_external_id: External ID of the SHOP case
            limit: Number of results to return, -1 for all results
        """
        result_list: ShopResultList = self._po.shop_based_day_ahead_bid_process.shop_result.list(
            case=case_external_id, limit=limit
        )
        return result_list

    def retrieve_shop_result(self, result_external_id: str) -> ShopResult:
        """Retrieve a shop result from CDF"""
        return self._po.shop_based_day_ahead_bid_process.shop_result.retrieve(external_id=result_external_id)

    def list_shop_versions(self) -> list[str]:
        """List the available version of SHOP in CDF. Does not include versions that are "local" to CShaaS.
        SHOP releases should have the following format:
        'SHOP-{VERSION}-pyshop-python{py_version}.linux.zip' and the metadata {'shop:type': 'shop-release'}.
        """
        # todo? Add an endpoint to list the available versions of SHOP via powerops API?
        return [file.name for file in self._cdf.files.list(metadata={"shop:type": "shop-release"}, limit=-1)]

    def retrieve_shop_case_graphql(self, case_external_id: str) -> ShopCase:
        """Retrieve a shop case from CDF using GraphQL.
        Instead of the external of IDs of ShopScenario, ShopModel and ShopFiles, the actual instances are fetched.
        This is provided as a fallback option for cases when `retrieve_shop_case` does not perform well."""
        graphql_response = self._po.shop_based_day_ahead_bid_process.graphql_query(_shop_case_query(case_external_id))
        if graphql_response:
            shop_case: ShopCase = graphql_response[0].as_read()
        else:
            raise ValueError(f"Failed to fetch ShopCase instance with external_id: {case_external_id}.")
        return shop_case

    def list_shop_results_graphql(self, case_external_id: str, limit: int) -> list[ShopResult]:
        """Retrieve a shop result, based on a case external id using GraphQL.
        This is provided as a fallback option for cases when `list_shop_results_for_case` does not perform well"""

        graphql_response = self._po.shop_based_day_ahead_bid_process.graphql_query(
            _shop_result_query(case_external_id, limit=limit)
        )

        if graphql_response is not None:
            shop_results: list[ShopResult] = [item.as_read() for item in graphql_response]
        else:
            raise ValueError(f"Failed to fetch ShopResult referencing ShopCase with external_id: {case_external_id}.")
        return shop_results


# Helper function for rendering graphql queries


def _shop_case_query(external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> str:
    """Render a GraphQL query to fetch a ShopCase instance by external_id."""
    query_template = """
    query RetrieveShopCaseByExternalId {{
      getShopCaseById(instance: {{ space: "{space}", externalId: "{external_id}" }}) {{
        items {{
          __typename
          createdTime
          endTime
          externalId
          lastUpdatedTime
          space
          startTime
          scenario {{
            createdTime
            externalId
            lastUpdatedTime
            name
            source
            space
            outputDefinition {{
              items {{
                attributeName
                createdTime
                externalId
                isStep
                lastUpdatedTime
                name
                objectName
                objectType
                space
                unit
              }}
            }}
            model {{
              createdTime
              externalId
              lastUpdatedTime
              modelVersion
              name
              shopVersion
              space
            }}
          }}
          shopFiles (sort: {{order: ASC}}) {{
            items {{
              createdTime
              externalId
              fileReferencePrefix
              isAscii
              label
              lastUpdatedTime
              name
              order
              space
              fileReference {{
                externalId
              }}
            }}
          }}
        }}
      }}
    }}
    """
    return query_template.format(space=space, external_id=external_id)


def _shop_result_query(case_external_id: str, limit: int) -> str:
    """Render a GraphQL query to fetch a ShopResult instance by case external id."""
    query_template = """
    query RetrieveShopResultByCaseExternalId {{
        listShopResult(
        filter: {{case: {{externalId: {{eq: "{case_external_id}"}} }} }}
        first: {limit}
    ) {{
        items {{
        __typename
        space
        createdTime
        externalId
        lastUpdatedTime
        preRun {{
            externalId
            id
            uploadedTime
            name
            metadata
            labels
        }}
        postRun {{
            externalId
            id
            labels
            uploadedTime
            name
            metadata
        }}
        messages {{
            externalId
            id
            labels
            uploadedTime
            name
            metadata
        }}
        cplexLogs {{
            externalId
            id
            labels
            uploadedTime
            name
            metadata
        }}
        case {{
            externalId
            lastUpdatedTime
            createdTime
        }}
        alerts {{
            items {{
            externalId
            lastUpdatedTime
            description
            createdTime
            alertType
            title
            time
            statusCode
            }}
        }}

        }}
    }}
    }}
    """
    return query_template.format(case_external_id=case_external_id, limit=limit)



from __future__ import annotations

import warnings
from typing import Any, ClassVar, Literal, no_type_check, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    TimeSeries,
)


__all__ = [
    "BenchmarkingProductionObligationDayAhead",
    "BenchmarkingProductionObligationDayAheadWrite",
    "BenchmarkingProductionObligationDayAheadApply",
    "BenchmarkingProductionObligationDayAheadList",
    "BenchmarkingProductionObligationDayAheadWriteList",
    "BenchmarkingProductionObligationDayAheadApplyList",
    "BenchmarkingProductionObligationDayAheadFields",
    "BenchmarkingProductionObligationDayAheadTextFields",
    "BenchmarkingProductionObligationDayAheadGraphQL",
]


BenchmarkingProductionObligationDayAheadTextFields = Literal["time_series", "name"]
BenchmarkingProductionObligationDayAheadFields = Literal["time_series", "name"]

_BENCHMARKINGPRODUCTIONOBLIGATIONDAYAHEAD_PROPERTIES_BY_FIELD = {
    "time_series": "timeSeries",
    "name": "name",
}

class BenchmarkingProductionObligationDayAheadGraphQL(GraphQLCore):
    """This represents the reading version of benchmarking production obligation day ahead, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking production obligation day ahead.
        data_record: The data record of the benchmarking production obligation day ahead node.
        time_series: The time series of the day ahead production obligation for benchmarking
        name: The name of the day ahead production obligation for benchmarking
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "BenchmarkingProductionObligationDayAhead", "1")
    time_series: Union[TimeSeries, dict, None] = Field(None, alias="timeSeries")
    name: Optional[str] = None

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> BenchmarkingProductionObligationDayAhead:
        """Convert this GraphQL format of benchmarking production obligation day ahead to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return BenchmarkingProductionObligationDayAhead(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            time_series=self.time_series,
            name=self.name,
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> BenchmarkingProductionObligationDayAheadWrite:
        """Convert this GraphQL format of benchmarking production obligation day ahead to the writing format."""
        return BenchmarkingProductionObligationDayAheadWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            time_series=self.time_series,
            name=self.name,
        )


class BenchmarkingProductionObligationDayAhead(DomainModel):
    """This represents the reading version of benchmarking production obligation day ahead.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking production obligation day ahead.
        data_record: The data record of the benchmarking production obligation day ahead node.
        time_series: The time series of the day ahead production obligation for benchmarking
        name: The name of the day ahead production obligation for benchmarking
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "BenchmarkingProductionObligationDayAhead", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "BenchmarkingProductionObligationDayAhead")
    time_series: Union[TimeSeries, str, None] = Field(None, alias="timeSeries")
    name: Optional[str] = None

    def as_write(self) -> BenchmarkingProductionObligationDayAheadWrite:
        """Convert this read version of benchmarking production obligation day ahead to the writing version."""
        return BenchmarkingProductionObligationDayAheadWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            time_series=self.time_series,
            name=self.name,
        )

    def as_apply(self) -> BenchmarkingProductionObligationDayAheadWrite:
        """Convert this read version of benchmarking production obligation day ahead to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BenchmarkingProductionObligationDayAheadWrite(DomainModelWrite):
    """This represents the writing version of benchmarking production obligation day ahead.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking production obligation day ahead.
        data_record: The data record of the benchmarking production obligation day ahead node.
        time_series: The time series of the day ahead production obligation for benchmarking
        name: The name of the day ahead production obligation for benchmarking
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "BenchmarkingProductionObligationDayAhead", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "BenchmarkingProductionObligationDayAhead")
    time_series: Union[TimeSeries, str, None] = Field(None, alias="timeSeries")
    name: Optional[str] = None

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.time_series is not None or write_none:
            properties["timeSeries"] = self.time_series if isinstance(self.time_series, str) or self.time_series is None else self.time_series.external_id

        if self.name is not None or write_none:
            properties["name"] = self.name


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        if isinstance(self.time_series, CogniteTimeSeries):
            resources.time_series.append(self.time_series)

        return resources


class BenchmarkingProductionObligationDayAheadApply(BenchmarkingProductionObligationDayAheadWrite):
    def __new__(cls, *args, **kwargs) -> BenchmarkingProductionObligationDayAheadApply:
        warnings.warn(
            "BenchmarkingProductionObligationDayAheadApply is deprecated and will be removed in v1.0. Use BenchmarkingProductionObligationDayAheadWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BenchmarkingProductionObligationDayAhead.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BenchmarkingProductionObligationDayAheadList(DomainModelList[BenchmarkingProductionObligationDayAhead]):
    """List of benchmarking production obligation day aheads in the read version."""

    _INSTANCE = BenchmarkingProductionObligationDayAhead

    def as_write(self) -> BenchmarkingProductionObligationDayAheadWriteList:
        """Convert these read versions of benchmarking production obligation day ahead to the writing versions."""
        return BenchmarkingProductionObligationDayAheadWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BenchmarkingProductionObligationDayAheadWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BenchmarkingProductionObligationDayAheadWriteList(DomainModelWriteList[BenchmarkingProductionObligationDayAheadWrite]):
    """List of benchmarking production obligation day aheads in the writing version."""

    _INSTANCE = BenchmarkingProductionObligationDayAheadWrite

class BenchmarkingProductionObligationDayAheadApplyList(BenchmarkingProductionObligationDayAheadWriteList): ...



def _create_benchmarking_production_obligation_day_ahead_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, ClassVar, Literal,  no_type_check, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)
from ._function_output import FunctionOutput, FunctionOutputWrite

if TYPE_CHECKING:
    from ._alert import Alert, AlertGraphQL, AlertWrite
    from ._bid_document_day_ahead import BidDocumentDayAhead, BidDocumentDayAheadGraphQL, BidDocumentDayAheadWrite
    from ._total_bid_matrix_calculation_input import TotalBidMatrixCalculationInput, TotalBidMatrixCalculationInputGraphQL, TotalBidMatrixCalculationInputWrite


__all__ = [
    "TotalBidMatrixCalculationOutput",
    "TotalBidMatrixCalculationOutputWrite",
    "TotalBidMatrixCalculationOutputApply",
    "TotalBidMatrixCalculationOutputList",
    "TotalBidMatrixCalculationOutputWriteList",
    "TotalBidMatrixCalculationOutputApplyList",
    "TotalBidMatrixCalculationOutputFields",
    "TotalBidMatrixCalculationOutputTextFields",
    "TotalBidMatrixCalculationOutputGraphQL",
]


TotalBidMatrixCalculationOutputTextFields = Literal["workflow_execution_id", "function_name", "function_call_id"]
TotalBidMatrixCalculationOutputFields = Literal["workflow_execution_id", "workflow_step", "function_name", "function_call_id"]

_TOTALBIDMATRIXCALCULATIONOUTPUT_PROPERTIES_BY_FIELD = {
    "workflow_execution_id": "workflowExecutionId",
    "workflow_step": "workflowStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
}

class TotalBidMatrixCalculationOutputGraphQL(GraphQLCore):
    """This represents the reading version of total bid matrix calculation output, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the total bid matrix calculation output.
        data_record: The data record of the total bid matrix calculation output node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        function_input: The function input field.
        alerts: An array of calculation level Alerts.
        bid_document: The bid document field.
    """
    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "TotalBidMatrixCalculationOutput", "1")
    workflow_execution_id: Optional[str] = Field(None, alias="workflowExecutionId")
    workflow_step: Optional[int] = Field(None, alias="workflowStep")
    function_name: Optional[str] = Field(None, alias="functionName")
    function_call_id: Optional[str] = Field(None, alias="functionCallId")
    function_input: Optional[TotalBidMatrixCalculationInputGraphQL] = Field(default=None, repr=False, alias="functionInput")
    alerts: Optional[list[AlertGraphQL]] = Field(default=None, repr=False)
    bid_document: Optional[BidDocumentDayAheadGraphQL] = Field(default=None, repr=False, alias="bidDocument")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values
    @field_validator("function_input", "alerts", "bid_document", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_read(self) -> TotalBidMatrixCalculationOutput:
        """Convert this GraphQL format of total bid matrix calculation output to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return TotalBidMatrixCalculationOutput(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            function_input=self.function_input.as_read() if isinstance(self.function_input, GraphQLCore) else self.function_input,
            alerts=[alert.as_read() for alert in self.alerts or []],
            bid_document=self.bid_document.as_read() if isinstance(self.bid_document, GraphQLCore) else self.bid_document,
        )


    # We do the ignore argument type as we let pydantic handle the type checking
    @no_type_check
    def as_write(self) -> TotalBidMatrixCalculationOutputWrite:
        """Convert this GraphQL format of total bid matrix calculation output to the writing format."""
        return TotalBidMatrixCalculationOutputWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            function_input=self.function_input.as_write() if isinstance(self.function_input, GraphQLCore) else self.function_input,
            alerts=[alert.as_write() for alert in self.alerts or []],
            bid_document=self.bid_document.as_write() if isinstance(self.bid_document, GraphQLCore) else self.bid_document,
        )


class TotalBidMatrixCalculationOutput(FunctionOutput):
    """This represents the reading version of total bid matrix calculation output.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the total bid matrix calculation output.
        data_record: The data record of the total bid matrix calculation output node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        function_input: The function input field.
        alerts: An array of calculation level Alerts.
        bid_document: The bid document field.
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "TotalBidMatrixCalculationOutput", "1")

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "TotalBidMatrixCalculationOutput")
    bid_document: Union[BidDocumentDayAhead, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidDocument")

    def as_write(self) -> TotalBidMatrixCalculationOutputWrite:
        """Convert this read version of total bid matrix calculation output to the writing version."""
        return TotalBidMatrixCalculationOutputWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            function_input=self.function_input.as_write() if isinstance(self.function_input, DomainModel) else self.function_input,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            bid_document=self.bid_document.as_write() if isinstance(self.bid_document, DomainModel) else self.bid_document,
        )

    def as_apply(self) -> TotalBidMatrixCalculationOutputWrite:
        """Convert this read version of total bid matrix calculation output to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class TotalBidMatrixCalculationOutputWrite(FunctionOutputWrite):
    """This represents the writing version of total bid matrix calculation output.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the total bid matrix calculation output.
        data_record: The data record of the total bid matrix calculation output node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        function_input: The function input field.
        alerts: An array of calculation level Alerts.
        bid_document: The bid document field.
    """
    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "TotalBidMatrixCalculationOutput", "1")

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "TotalBidMatrixCalculationOutput")
    bid_document: Union[BidDocumentDayAheadWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidDocument")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        properties: dict[str, Any] = {}

        if self.workflow_execution_id is not None:
            properties["workflowExecutionId"] = self.workflow_execution_id

        if self.workflow_step is not None:
            properties["workflowStep"] = self.workflow_step

        if self.function_name is not None:
            properties["functionName"] = self.function_name

        if self.function_call_id is not None:
            properties["functionCallId"] = self.function_call_id

        if self.function_input is not None:
            properties["functionInput"] = {
                "space":  self.space if isinstance(self.function_input, str) else self.function_input.space,
                "externalId": self.function_input if isinstance(self.function_input, str) else self.function_input.external_id,
            }

        if self.bid_document is not None:
            properties["bidDocument"] = {
                "space":  self.space if isinstance(self.bid_document, str) else self.bid_document.space,
                "externalId": self.bid_document if isinstance(self.bid_document, str) else self.bid_document.external_id,
            }


        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=self._view_id,
                        properties=properties,
                )],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())



        edge_type = dm.DirectRelationReference("power_ops_types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=alert,
                edge_type=edge_type,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.function_input, DomainModelWrite):
            other_resources = self.function_input._to_instances_write(cache)
            resources.extend(other_resources)

        if isinstance(self.bid_document, DomainModelWrite):
            other_resources = self.bid_document._to_instances_write(cache)
            resources.extend(other_resources)

        return resources


class TotalBidMatrixCalculationOutputApply(TotalBidMatrixCalculationOutputWrite):
    def __new__(cls, *args, **kwargs) -> TotalBidMatrixCalculationOutputApply:
        warnings.warn(
            "TotalBidMatrixCalculationOutputApply is deprecated and will be removed in v1.0. Use TotalBidMatrixCalculationOutputWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "TotalBidMatrixCalculationOutput.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class TotalBidMatrixCalculationOutputList(DomainModelList[TotalBidMatrixCalculationOutput]):
    """List of total bid matrix calculation outputs in the read version."""

    _INSTANCE = TotalBidMatrixCalculationOutput

    def as_write(self) -> TotalBidMatrixCalculationOutputWriteList:
        """Convert these read versions of total bid matrix calculation output to the writing versions."""
        return TotalBidMatrixCalculationOutputWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> TotalBidMatrixCalculationOutputWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class TotalBidMatrixCalculationOutputWriteList(DomainModelWriteList[TotalBidMatrixCalculationOutputWrite]):
    """List of total bid matrix calculation outputs in the writing version."""

    _INSTANCE = TotalBidMatrixCalculationOutputWrite

class TotalBidMatrixCalculationOutputApplyList(TotalBidMatrixCalculationOutputWriteList): ...



def _create_total_bid_matrix_calculation_output_filter(
    view_id: dm.ViewId,
    workflow_execution_id: str | list[str] | None = None,
    workflow_execution_id_prefix: str | None = None,
    min_workflow_step: int | None = None,
    max_workflow_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    function_input: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    bid_document: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(workflow_execution_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id))
    if workflow_execution_id and isinstance(workflow_execution_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("workflowExecutionId"), values=workflow_execution_id))
    if workflow_execution_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id_prefix))
    if min_workflow_step is not None or max_workflow_step is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("workflowStep"), gte=min_workflow_step, lte=max_workflow_step))
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if function_input and isinstance(function_input, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionInput"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": function_input}))
    if function_input and isinstance(function_input, tuple):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionInput"), value={"space": function_input[0], "externalId": function_input[1]}))
    if function_input and isinstance(function_input, list) and isinstance(function_input[0], str):
        filters.append(dm.filters.In(view_id.as_property_ref("functionInput"), values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in function_input]))
    if function_input and isinstance(function_input, list) and isinstance(function_input[0], tuple):
        filters.append(dm.filters.In(view_id.as_property_ref("functionInput"), values=[{"space": item[0], "externalId": item[1]} for item in function_input]))
    if bid_document and isinstance(bid_document, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("bidDocument"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": bid_document}))
    if bid_document and isinstance(bid_document, tuple):
        filters.append(dm.filters.Equals(view_id.as_property_ref("bidDocument"), value={"space": bid_document[0], "externalId": bid_document[1]}))
    if bid_document and isinstance(bid_document, list) and isinstance(bid_document[0], str):
        filters.append(dm.filters.In(view_id.as_property_ref("bidDocument"), values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in bid_document]))
    if bid_document and isinstance(bid_document, list) and isinstance(bid_document[0], tuple):
        filters.append(dm.filters.In(view_id.as_property_ref("bidDocument"), values=[{"space": item[0], "externalId": item[1]} for item in bid_document]))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import logging
from typing import Optional, Union, cast

import numpy as np
import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import DatapointsList, Event, LabelFilter, RelationshipList
from cognite.client.exceptions import CogniteDuplicatedError
from cognite.client.utils import ms_to_datetime
from cognite.client.utils.useful_types import SequenceNotStr

from cognite.powerops.utils.require import require

logger = logging.getLogger(__name__)


def retrieve_event(client: CogniteClient, external_id: str) -> Event:
    event = client.events.retrieve(external_id=external_id)
    if event is None:
        raise ValueError(f"Event not found: {external_id}")
    return event


def remove_duplicates(lst: list) -> list:
    return list(set(lst))


def _overlapping_keys(dict_: dict, other: dict) -> list[str]:
    return list(set(dict_.keys()).intersection(other.keys()))


def merge_dicts(*args: dict) -> dict:
    merged: dict = {}
    for dict_ in args:
        assert isinstance(dict_, dict), f"Expected dict, got {type(dict_)}!"
        overlap = _overlapping_keys(dict_, merged)
        if overlap:
            raise Exception(f"Key collision on '{overlap}' when merging dictionaries!")
        merged = {**merged, **dict_}
    return merged


def retrieve_relationships_from_source_ext_id(
    client: CogniteClient,
    source_ext_id: str,
    label_ext_id: Optional[Union[str, list[str]]],
    target_types: SequenceNotStr[str] | None = None,
) -> RelationshipList:
    """
    Retrieve relationships between source and target using the source ext id.
    Using the `containsAny` filter, we can retrieve all relationships with  given label.
    """
    if isinstance(label_ext_id, str):
        label_ext_id = [label_ext_id]
    if label_ext_id is not None:
        _labels = LabelFilter(contains_any=label_ext_id)
    else:
        _labels = None
    return client.relationships.list(
        source_external_ids=[source_ext_id], labels=_labels, limit=-1, target_types=target_types
    )


def _retrieve_range(client: CogniteClient, external_ids: list[str], start: int, end: int) -> pd.DataFrame:
    # TODO: Upgrade cognite-sdk to v5 (or later), and see how much of the code we can replace with direct SDK calls
    # - client.time_series.data.retrieve_dataframe(…, uniform_index=True) should give us almost what we want,
    # but maybe we need to be careful with cases where there is more than 1 hour between values
    # (I do not remember if this is an issue only for some aggregates like average, or for all).
    # And maybe we need to parametrise the "minimum resolution"
    # (seems to assume 1 hour, but we should support sub-hourly resolution)
    # Retrieve raw datapoints
    external_ids = remove_duplicates(external_ids)
    if not external_ids:
        return pd.DataFrame()

    start_dt = ms_to_datetime(start).replace(tzinfo=None)  # UTC implied
    end_dt = ms_to_datetime(end).replace(tzinfo=None)  # UTC implied
    logger.debug(f"Retrieving {external_ids} between '{start_dt}' and '{end_dt}'")
    df_range = client.time_series.data.retrieve(  # type: ignore[union-attr]
        external_id=external_ids, start=start, end=end, ignore_unknown_ids=True
    ).to_pandas()

    # Retrieve latest datapoints before start
    df_latest = client.time_series.data.retrieve_latest(  # type: ignore[union-attr]
        external_id=external_ids, before=start, ignore_unknown_ids=True
    ).to_pandas()

    # Make sure we have a start timestamp in range
    if df_range.empty:
        df_range = pd.DataFrame(
            columns=df_latest.columns,
            index=pd.DatetimeIndex(data=np.array([int(start)], dtype="datetime64[ms]")),
            dtype=float,
        )

    # Add the latest datapoints to the DataFrame
    df_raw = df_range.combine_first(df_latest)

    # Must retrieve time series metadata to correctly resample and aggregate datapoints
    time_series = client.time_series.retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=True)
    step_columns = [ts.external_id for ts in time_series if ts.is_step]
    linear_columns = [ts.external_id for ts in time_series if not ts.is_step]
    logger.debug(f"time_series.is_step: True [{len(step_columns)}] False [{len(linear_columns)}]")

    # Step interpolation of time series with .is_step=False
    # NOTE: do not need to upsample when forward filling
    # TODO: note 2x ffill()
    df_step = df_raw[step_columns].ffill().resample("1h").ffill()

    # Linear interpolation of time series with .is_step=False
    # TODO: must upsample before downsampling?
    # TODO: confirm operations
    intermediate_df = df_raw[linear_columns].resample("1min").interpolate()
    df_linear = intermediate_df.resample("1h").interpolate()

    # Merge the step interpolated and linearly interpolated DataFrames
    df_combined = df_step.combine_first(df_linear)

    # Only return datapoints within the range
    df_filtered = df_combined[start_dt:end_dt]  # type: ignore[misc]

    return df_filtered


def retrieve_range(client: CogniteClient, external_ids: list[str], start: int, end: int) -> dict[str, pd.Series]:
    retrieved_range_df = _retrieve_range(client=client, external_ids=external_ids, start=start, end=end)
    return {col: retrieved_range_df[col].dropna() for col in retrieved_range_df.columns}


# TODO: refactor
# TODO: naming: time_series vs. datapoints
def retrieve_latest(client: CogniteClient, external_ids: list[Optional[str]], before: int) -> dict[str, pd.Series]:
    if not external_ids:
        return {}
    external_ids = remove_duplicates(external_ids)
    logger.debug(f"Retrieving {external_ids} before '{ms_to_datetime(before)}'")
    time_series = cast(
        DatapointsList,
        client.time_series.data.retrieve_latest(external_id=external_ids, before=before, ignore_unknown_ids=True),
    )

    # For (Cog)Datapoints in (Cog)DatapointsList
    for datapoints in time_series:
        if len(datapoints) > 0:  # TODO: what to do about ts with no datapoints?
            datapoints.timestamp[0] = before

    res = {
        require(datapoints.external_id): datapoints.to_pandas().iloc[:, 0]  # iloc to convert DataFrame to Series
        for datapoints in time_series
        if len(datapoints) > 0
    }
    if missing := set(external_ids).difference(res):
        logger.warning(f"Missing: {', '.join(map(str, missing))}")
    return res


def retrieve_time_series_datapoints(  # type: ignore[no-untyped-def]
    client: CogniteClient, mappings, start: int, end: int
) -> dict[str, pd.Series]:
    time_series_start = retrieve_latest(
        client=client,
        external_ids=[mapping.cdf_time_series for mapping in mappings if mapping.retrieve == "START"],
        before=start,
    )
    time_series_end = retrieve_latest(
        client=client,
        external_ids=[mapping.cdf_time_series for mapping in mappings if mapping.retrieve == "END"],
        before=end,
    )
    time_series_range = retrieve_range(
        client=client,
        external_ids=[mapping.cdf_time_series for mapping in mappings if mapping.retrieve == "RANGE"],
        start=start,
        end=end,
    )
    _time_series_none = [
        ".".join(filter(None, [mapping.shop_object_type, mapping.shop_object_name, mapping.shop_attribute_name]))
        for mapping in mappings
        if not mapping.retrieve
    ]
    logger.debug(f"Not retrieving datapoints for {_time_series_none}")

    return merge_dicts(time_series_start, time_series_end, time_series_range)


def create_event(client: CogniteClient, event: Event) -> None:
    try:
        client.events.create(event)
    except CogniteDuplicatedError:
        logger.warning(f"Event with external_id '{event.external_id}' already exists")
        exit(1)
